import unittest
import re
import time
import warnings
from datetime import datetime, timedelta

from robot.utils.asserts import (assert_equal, assert_raises_with_msg,
                                 assert_true, assert_not_none)

from robot.utils.robottime import (timestr_to_secs, secs_to_timestr, get_time,
                                   parse_time, format_time, get_elapsed_time,
                                   get_timestamp, timestamp_to_secs,
                                   elapsed_time_to_string, _get_timetuple)


EXAMPLE_TIME = time.mktime(datetime(2007, 9, 20, 16, 15, 14).timetuple())


class TestTime(unittest.TestCase):

    def test_get_timetuple_excluding_millis(self):
        assert_equal(_get_timetuple(12345)[:-1], time.localtime(12345)[:6])

    def test_get_current_timetuple_excluding_millis(self):
        while True:
            expected = time.localtime(time.time())
            actual = _get_timetuple()
            # make sure got same times and _get_timetuple() did not round millis
            if expected == time.localtime(time.time()) and actual[-1] > 0:
                break
        assert_equal(actual[:-1], expected[:6])

    def test_get_timetuple_millis(self):
        assert_equal(_get_timetuple(12345)[-2:], (45, 0))
        assert_equal(_get_timetuple(12345.12345)[-2:], (45, 123))
        assert_equal(_get_timetuple(12345.67890)[-2:], (45, 679))
        assert_equal(_get_timetuple(12345.99999)[-2:], (46, 0))

    def test_timestr_to_secs_with_numbers(self):
        for inp, exp in [(1, 1),
                         (42, 42),
                         (1.1, 1.1),
                         (3.142, 3.142),
                         (-1, -1),
                         (-1.1, -1.1),
                         (0, 0),
                         (0.55555, 0.556),
                         (11.111111, 11.111),
                         ('1e2', 100),
                         ('-1.5e3', -1500)]:
            assert_equal(timestr_to_secs(inp), exp, inp)
            if not isinstance(inp, str):
                assert_equal(timestr_to_secs(str(inp)), exp, inp)

    def test_timestr_to_secs_uses_bankers_rounding(self):
        assert_equal(timestr_to_secs(0.1, 0), 0)
        assert_equal(timestr_to_secs(0.5, 0), 0)
        assert_equal(timestr_to_secs(0.51, 0), 1)
        assert_equal(timestr_to_secs(0.99, 0), 1)
        assert_equal(timestr_to_secs(1.0, 0), 1)
        assert_equal(timestr_to_secs(1.499, 0), 1)
        assert_equal(timestr_to_secs(1.5, 0), 2)

    def test_timestr_to_secs_with_time_string(self):
        for inp, exp in [('1s', 1),
                         ('0 day 1 MINUTE 2 S 42 millis', 62.042),
                         ('1minute 0sec 10 millis', 60.01),
                         ('9 9 secs    5  3 4 m i l l i s e co n d s', 99.534),
                         ('10DAY10H10M10SEC', 900610),
                         ('1day 23h 46min 7s 666ms', 171967.666),
                         ('1.5min 1.5s', 91.5),
                         ('1.5 days', 60*60*36),
                         ('1 day', 60*60*24),
                         ('2 days', 2*60*60*24),
                         ('1 d', 60*60*24),
                         ('1 hour', 60*60),
                         ('3 hours', 3*60*60),
                         ('1 h', 60*60),
                         ('1 minute', 60),
                         ('2 minutes', 2*60),
                         ('1 min', 60),
                         ('2 mins', 2*60),
                         ('1 m', 60),
                         ('1 second', 1),
                         ('2 seconds', 2),
                         ('1 sec', 1),
                         ('2 secs', 2),
                         ('1 s', 1),
                         ('1 millisecond', 0.001),
                         ('2 milliseconds', 0.002),
                         ('1 millisec', 0.001),
                         ('2 millisecs', 0.002),
                         ('1234 millis', 1.234),
                         ('1 msec', 0.001),
                         ('2 msecs', 0.002),
                         ('1 ms', 0.001),
                         ('-1s', -1),
                         ('- 1 min 2 s', -62),
                         ('0.1millis', 0),
                         ('0.5ms', 0.001),
                         ('0day 0hour 0minute 0seconds 0millisecond', 0)]:
            assert_equal(timestr_to_secs(inp), exp, inp)

    def test_timestr_to_secs_with_time_string_ns_accuracy(self):
        for input, expected in [("1 us", 1E-6),
                                ("1 μs", 1E-6),
                                ("1 microsecond", 1E-6),
                                ("1 microseconds", 1E-6),
                                ("2 us", 2E-6),
                                ("1 ns", 1E-9),
                                ("1 nanosecond", 1E-9),
                                ("1 nanoseconds", 1E-9),
                                ("2 ns", 2E-9),
                                ("-100 ns", -100E-9),
                                ("1.2us", 1.2E-6)]:
            assert_equal(timestr_to_secs(input, round_to=9), expected)

    def test_timestr_to_secs_with_timer_string(self):
        for inp, exp in [('00:00:00', 0),
                         ('00:00:01', 1),
                         ('01:02:03', 3600 + 2*60 + 3),
                         ('100:00:00', 100*3600),
                         ('1:00:00', 3600),
                         ('11:00:00', 11*3600),
                         ('00:00', 0),
                         ('00:01', 1),
                         ('42:01', 42*60 + 1),
                         ('100:00', 100*60),
                         ('100:100', 100*60 + 100),
                         ('100:100:100', 100*3600 + 100*60 + 100),
                         ('1:1:1', 3600 + 60 + 1),
                         ('0001:0001:0001', 3600 + 60 + 1),
                         ('-00:00:00', 0),
                         ('-00:01:10', -70),
                         ('-1:2:3', -3600 - 2*60 - 3),
                         ('+00:00:00', 0),
                         ('+00:01:10', 70),
                         ('+1:2:3', 3600 + 2*60 + 3),
                         ('00:00:00.0', 0),
                         ('00:00:00.000', 0),
                         ('00:00:00.000000000', 0),
                         ('00:00:00.1', 0.1),
                         ('00:00:00.42', 0.42),
                         ('00:00:00.001', 0.001),
                         ('00:00:00.123', 0.123),
                         ('00:00:00.1234', 0.123),
                         ('00:00:00.12345', 0.123),
                         ('00:00:00.12356', 0.124),
                         ('00:00:00.999', 0.999),
                         ('00:00:00.9995001', 1),
                         ('00:00:00.000000001', 0)]:
            assert_equal(timestr_to_secs(inp), exp, inp)
            if '.' not in inp:
                inp += '.500'
                exp += 0.5 if inp[0] != '-' else -0.5
                assert_equal(timestr_to_secs(inp), exp, inp)

    def test_timestr_to_secs_with_timedelta(self):
        assert_equal(timestr_to_secs(timedelta(minutes=1)), 60)
        assert_equal(timestr_to_secs(timedelta(microseconds=1000)), 0.001)

    def test_timestr_to_secs_custom_rounding(self):
        secs = 0.123456789
        for round_to in 0, 1, 6:
            expected = round(secs, round_to)
            assert_equal(timestr_to_secs(secs, round_to), expected)
            assert_equal(timestr_to_secs(str(secs), round_to), expected)

    def test_timestr_to_secs_no_rounding(self):
        secs = 0.123456789
        assert_equal(timestr_to_secs(secs, round_to=None), secs)
        assert_equal(timestr_to_secs(str(secs), round_to=None), secs)

    def test_timestr_to_secs_with_invalid(self):
        for inv in ['', 'foo', 'foo days', '1sec 42 millis 3', '1min 2w',
                    '01:02:03:04', '01:02:03foo', 'foo01:02:03', None]:
            assert_raises_with_msg(ValueError, "Invalid time string '%s'." % inv,
                                   timestr_to_secs, inv)

    def test_timestr_to_secs_accept_plain_values(self):
        with warnings.catch_warnings(record=True) as w:
            assert_raises_with_msg(ValueError, "Invalid time string '100'.",
                                   timestr_to_secs, '100', accept_plain_values=False)
            assert_equal(str(w[-1].message),
                         "'accept_plain_values' is deprecated and will be removed in RF 7.0.")

    def test_secs_to_timestr(self):
        for inp, compact, verbose in [
            (0.001, '1ms', '1 millisecond'),
            (0.002, '2ms', '2 milliseconds'),
            (0.9999, '1s', '1 second'),
            (1, '1s', '1 second'),
            (1.9999, '2s', '2 seconds'),
            (2, '2s', '2 seconds'),
            (60, '1min', '1 minute'),
            (120, '2min', '2 minutes'),
            (3600, '1h', '1 hour'),
            (7200, '2h', '2 hours'),
            (60*60*24, '1d', '1 day'),
            (60*60*48, '2d', '2 days'),
            (171967.667, '1d 23h 46min 7s 667ms',
             '1 day 23 hours 46 minutes 7 seconds 667 milliseconds'),
            (7320, '2h 2min', '2 hours 2 minutes'),
            (7210.05, '2h 10s 50ms', '2 hours 10 seconds 50 milliseconds') ,
            (11.1111111, '11s 111ms', '11 seconds 111 milliseconds'),
            (0.55555555, '556ms', '556 milliseconds'),
            (0, '0s', '0 seconds'),
            (9999.9999, '2h 46min 40s', '2 hours 46 minutes 40 seconds'),
            (10000, '2h 46min 40s', '2 hours 46 minutes 40 seconds'),
            (-1, '- 1s', '- 1 second'),
            (-171967.667, '- 1d 23h 46min 7s 667ms',
             '- 1 day 23 hours 46 minutes 7 seconds 667 milliseconds')]:
            assert_equal(secs_to_timestr(inp, compact=True), compact, inp)
            assert_equal(secs_to_timestr(inp), verbose, inp)

    def test_format_time(self):
        timetuple = (2005, 11, 2, 14, 23, 12, 123)
        for seps, exp in [(('-',' ',':'), '2005-11-02 14:23:12'),
                          (('', '-', ''), '20051102-142312'),
                          (('-',' ',':','.'), '2005-11-02 14:23:12.123')]:
            assert_equal(format_time(timetuple, *seps), exp)

    def test_get_timestamp(self):
        for seps, pattern in [
            ((), r'^\d{8} \d\d:\d\d:\d\d.\d\d\d$'),
            (('', ' ', ':', None), r'^\d{8} \d\d:\d\d:\d\d$'),
            (('', '', '', None), r'^\d{14}$'),
            (('-', '&nbsp;', ':', ';'), r'^\d{4}-\d\d-\d\d&nbsp;\d\d:\d\d:\d\d;\d\d\d$')
        ]:
            ts = get_timestamp(*seps)
            assert_not_none(re.search(pattern, ts),
                            "'%s' didn't match '%s'" % (ts, pattern), False)

    def test_timestamp_to_secs_with_default(self):
        assert_equal(timestamp_to_secs('20070920 16:15:14.123'), EXAMPLE_TIME+0.123)

    def test_timestamp_to_secs_with_seps(self):
        result = timestamp_to_secs('2007-09-20#16x15x14M123', ('-','#','x','M'))
        assert_equal(result, EXAMPLE_TIME+0.123)

    def test_timestamp_to_secs_with_millis(self):
        result = timestamp_to_secs('20070920 16:15:14.123')
        assert_equal(result, EXAMPLE_TIME+0.123)

    def test_get_elapsed_time(self):
        starttime = '20060526 14:01:10.500'
        for endtime, expected in [('20060526 14:01:10.500', 0),
                                  ('20060526 14:01:10.500',0),
                                  ('20060526 14:01:10.501', 1),
                                  ('20060526 14:01:10.777', 277),
                                  ('20060526 14:01:11.000', 500),
                                  ('20060526 14:01:11.321', 821),
                                  ('20060526 14:01:11.499', 999),
                                  ('20060526 14:01:11.500', 1000),
                                  ('20060526 14:01:11.501', 1001),
                                  ('20060526 14:01:11.000', 500),
                                  ('20060526 14:01:11.500', 1000),
                                  ('20060526 14:01:11.510', 1010),
                                  ('20060526 14:01:11.512',1012),
                                  ('20060601 14:01:10.499', 518399999),
                                  ('20060601 14:01:10.500', 518400000),
                                  ('20060601 14:01:10.501', 518400001)]:
            actual = get_elapsed_time(starttime, endtime)
            assert_equal(actual, expected, endtime)

    def test_get_elapsed_time_negative(self):
        starttime = '20060526 14:01:10.500'
        for endtime, expected in [('20060526 14:01:10.499', -1),
                                  ('20060526 14:01:10.000', -500),
                                  ('20060526 14:01:09.900', -600),
                                  ('20060526 14:01:09.501', -999),
                                  ('20060526 14:01:09.500', -1000),
                                  ('20060526 14:01:09.499', -1001)]:
            actual = get_elapsed_time(starttime, endtime)
            assert_equal(actual, expected, endtime)

    def test_elapsed_time_to_string(self):
        for elapsed, expected in [(0, '00:00:00.000'),
                                  (0.1, '00:00:00.000'),
                                  (0.5, '00:00:00.000'),
                                  (0.50001, '00:00:00.001'),
                                  (1, '00:00:00.001'),
                                  (1.5, '00:00:00.002'),
                                  (42, '00:00:00.042'),
                                  (999, '00:00:00.999'),
                                  (999.9, '00:00:01.000'),
                                  (1000, '00:00:01.000'),
                                  (1001, '00:00:01.001'),
                                  (60000, '00:01:00.000'),
                                  (600000, '00:10:00.000'),
                                  (654321, '00:10:54.321'),
                                  (660000, '00:11:00.000'),
                                  (3600000, '01:00:00.000'),
                                  (36000000, '10:00:00.000'),
                                  (360000000, '100:00:00.000'),
                                  (360000000 + 36000000 + 3600000 + 660000 + 11111,
                                   '111:11:11.111')]:
            assert_equal(elapsed_time_to_string(elapsed), expected, elapsed)
            if expected != '00:00:00.000':
                assert_equal(elapsed_time_to_string(-1 * elapsed),
                             '-' + expected, elapsed)

    def test_elapsed_time_to_string_without_millis(self):
        for elapsed, expected in [(0, '00:00:00'),
                                  (1, '00:00:00'),
                                  (500, '00:00:00'),
                                  (500.001, '00:00:01'),
                                  (999, '00:00:01'),
                                  (1000, '00:00:01'),
                                  (1499.999, '00:00:01'),
                                  (1500, '00:00:02'),
                                  (59499.9, '00:00:59'),
                                  (59500.0, '00:01:00'),
                                  (59999, '00:01:00'),
                                  (60000, '00:01:00'),
                                  (654321, '00:10:54'),
                                  (654500, '00:10:54'),
                                  (654501, '00:10:55'),
                                  (3599999, '01:00:00'),
                                  (3600000, '01:00:00'),
                                  (359999999, '100:00:00'),
                                  (360000000, '100:00:00'),
                                  (360000500, '100:00:00'),
                                  (360000500.001, '100:00:01')]:
            assert_equal(elapsed_time_to_string(elapsed, include_millis=False),
                         expected, elapsed)
            if expected != '00:00:00':
                assert_equal(elapsed_time_to_string(-1 * elapsed, False),
                             '-' + expected, elapsed)

    def test_parse_time_with_valid_times(self):
        for input, expected in [('100', 100),
                                ('2007-09-20 16:15:14', EXAMPLE_TIME),
                                ('20070920 161514', EXAMPLE_TIME)]:
            assert_equal(parse_time(input), expected)

    def test_parse_time_with_now_and_utc(self):
        for input, adjusted in [('now', 0),
                                ('NOW', 0),
                                ('Now', 0),
                                ('now+100seconds', 100),
                                ('now    -    100    seconds   ', -100),
                                ('now + 1 day 100 seconds', 86500),
                                ('now - 1 day 100 seconds', -86500),
                                ('now + 1day 10hours 1minute 10secs', 122470),
                                ('NOW - 1D 10H 1MIN 10S', -122470)]:
            now = int(time.time())
            parsed = parse_time(input)
            expected = now + adjusted
            if time.localtime(now).tm_isdst is not time.localtime(expected).tm_isdst:
                dst_diff = time.timezone - time.altzone
                expected += dst_diff if time.localtime(now).tm_isdst else -dst_diff
            assert_true(expected - parsed < 0.1)
            parsed = parse_time(input.upper().replace('NOW', 'UtC'))
            zone = time.altzone if time.localtime(now).tm_isdst else time.timezone
            expected += zone
            assert_true(expected - parsed < 0.1)

    def test_get_time_with_zero(self):
        assert_equal(get_time('epoch', 0), 0)

    def test_parse_modified_time_with_invalid_times(self):
        for value, msg in [("-100", "Epoch time must be positive (got -100)."),
                           ("YYYY-MM-DD hh:mm:ss",
                            "Invalid time format 'YYYY-MM-DD hh:mm:ss'."),
                           ("now + foo", "Invalid time string 'foo'."),
                           ("now -    2a ", "Invalid time string '2a'."),
                           ("now+", "Invalid time string ''."),
                           ("nowadays", "Invalid time format 'nowadays'.")]:
            assert_raises_with_msg(ValueError, msg, parse_time, value)

    def test_parse_time_and_get_time_must_round_seconds_down(self):
        # Rounding to closest second, instead of rounding down, could give
        # times that are greater then e.g. timestamps of files created
        # afterwards.
        self._verify_parse_time_and_get_time_rounding()
        time.sleep(0.5)
        self._verify_parse_time_and_get_time_rounding()

    def _verify_parse_time_and_get_time_rounding(self):
        secs = lambda: int(time.time()) % 60
        start_secs = secs()
        gt_result = get_time()[-2:]
        pt_result = parse_time('NOW') % 60
        # Check that seconds have not changed during test
        if secs() == start_secs:
            assert_equal(gt_result, '%02d' % start_secs)
            assert_equal(pt_result, start_secs)

    def test_get_timestamp_without_millis(self):
        # Need to test twice to verify also possible cached timestamp
        assert_true(re.match(r'\d{8} \d\d:\d\d:\d\d', get_timestamp(millissep=None)))
        assert_true(re.match(r'\d{8} \d\d:\d\d:\d\d', get_timestamp(millissep=None)))


if __name__ == "__main__":
    unittest.main()



import unittest

from robot.result import TestSuite
from robot.result.keywordremover import WaitUntilKeywordSucceedsRemover
from robot.utils.asserts import assert_equal


class TestWUKSRemover(unittest.TestCase):

    def test_empty(self):
        self._assert_removed()

    def test_one_passing(self):
        self._assert_removed(passing=1, expected=1)

    def test_one_failing(self):
        self._assert_removed(failing=1, expected=1)

    def test_failing_and_passing(self):
        self._assert_removed(failing=1, passing=1, expected=2)
        self._assert_removed(failing=9, passing=1, expected=2)

    def test_only_messages(self):
        self._assert_removed(messages=1, expected=1)
        self._assert_removed(messages=7, expected=7)

    def test_keywords_and_messages(self):
        self._assert_removed(passing=1, messages=1, expected=2)
        self._assert_removed(failing=1, messages=2, expected=3)
        self._assert_removed(failing=1, passing=1, messages=2, expected=4)
        self._assert_removed(failing=9, passing=1, messages=3, expected=5)

    def _assert_removed(self, failing=0, passing=0, messages=0, expected=0):
        suite = TestSuite()
        kw = suite.tests.create().body.create_keyword(
            libname='BuiltIn', kwname='Wait Until Keyword Succeeds'
        )
        for i in range(failing):
            kw.body.create_keyword(status='FAIL')
        for i in range(passing):
            kw.body.create_keyword(status='PASS')
        for i in range(messages):
            kw.body.create_message()
        suite.visit(WaitUntilKeywordSucceedsRemover())
        assert_equal(len(kw.body), expected)


if __name__ == '__main__':
    unittest.main()



from enum import Flag, Enum, IntFlag, IntEnum
from datetime import datetime, date, timedelta
from decimal import Decimal
from pathlib import Path, PurePath    # Path needed by `eval()` in `_validate_type()`.

from robot.api.deco import keyword


class MyEnum(Enum):
    FOO = 1
    bar = 'xxx'


class MyFlag(Flag):
    RED = 1
    BLUE = 2


class MyIntEnum(IntEnum):
    ON = 1
    OFF = 0


class MyIntFlag(IntFlag):
    R = 4
    W = 2
    X = 1


class Unknown:
    pass


def integer(argument=1, expected=None):
    _validate_type(argument, expected)


def float_(argument=-1.0, expected=None):
    _validate_type(argument, expected)


def decimal(argument=Decimal('1.2'), expected=None):
    _validate_type(argument, expected)


def boolean(argument=True, expected=None):
    _validate_type(argument, expected)


def string(argument='', expected=None):
    _validate_type(argument, expected)


def bytes_(argument=b'', expected=None):
    _validate_type(argument, expected)


def bytearray_(argument=bytearray(), expected=None):
    _validate_type(argument, expected)


def datetime_(argument=datetime.now(), expected=None):
    _validate_type(argument, expected)


def date_(argument=date.today(), expected=None):
    _validate_type(argument, expected)


def timedelta_(argument=timedelta(), expected=None):
    _validate_type(argument, expected)


def path(argument=Path(), expected=None):
    _validate_type(argument, expected)


def pure_path(argument=PurePath(), expected=None):
    _validate_type(argument, expected)


def enum(argument=MyEnum.FOO, expected=None):
    _validate_type(argument, expected)


def flag(argument=MyFlag.RED, expected=None):
    _validate_type(argument, expected)


def int_enum(argument=MyIntEnum.ON, expected=None):
    _validate_type(argument, expected)


def int_flag(argument=MyIntFlag.X, expected=None):
    _validate_type(argument, expected)


def none(argument=None, expected=None):
    _validate_type(argument, expected)


def list_(argument=['mutable', 'defaults', 'are', 'bad'], expected=None):
    _validate_type(argument, expected)


def tuple_(argument=('immutable', 'defaults', 'are', 'ok'), expected=None):
    _validate_type(argument, expected)


def dictionary(argument={'mutable defaults': 'are bad'}, expected=None):
    _validate_type(argument, expected)


def set_(argument={'mutable', 'defaults', 'are', 'bad'}, expected=None):
    _validate_type(argument, expected)


def frozenset_(argument=frozenset({'immutable', 'ok'}), expected=None):
    _validate_type(argument, expected)


def unknown(argument=Unknown(), expected=None):
    _validate_type(argument, expected)


def kwonly(*, argument=0.0, expected=None):
    _validate_type(argument, expected)


@keyword(types={'argument': timedelta})
def types_via_keyword_deco_override(argument=0, expected=None):
    _validate_type(argument, expected)


@keyword(name='None as types via @keyword disables', types=None)
def none_as_types(argument=0, expected=None):
    _validate_type(argument, expected)


@keyword(name="Empty types via @keyword doesn't override", types=[])
def empty_list_as_types(argument=0, expected=None):
    _validate_type(argument, expected)


@keyword(name="@keyword without types doesn't override")
def keyword_deco_alone_does_not_override(argument=0, expected=None):
    _validate_type(argument, expected)


def _validate_type(argument, expected):
    if isinstance(expected, str):
        expected = eval(expected)
    if argument != expected or type(argument) != type(expected):
        raise AssertionError('%r (%s) != %r (%s)'
                             % (argument, type(argument).__name__,
                                expected, type(expected).__name__))



from functools import wraps


def decorator(f):
    @wraps(f)
    def wrapper(*args, **kws):
        return f(*args, **kws)
    return wrapper


@decorator
def wrapped_function():
    pass


@decorator
def wrapped_function_with_arguments(a, b=2):
    pass



import unittest

from robot.utils.asserts import assert_equal
from robot.utils.encoding import console_decode, console_encode, CONSOLE_ENCODING


UNICODE = 'hyvä'
ENCODED = UNICODE.encode(CONSOLE_ENCODING)


class TestConsoleDecode(unittest.TestCase):

    def test_decode(self):
        assert_equal(console_decode(ENCODED), UNICODE)

    def test_unicode_is_returned_as_is(self):
        assert_equal(console_decode(UNICODE), UNICODE)


class TestConsoleEncode(unittest.TestCase):

    def test_unicode_is_returned_as_is_by_default(self):
        assert_equal(console_encode(UNICODE), UNICODE)

    def test_force_encoding(self):
        assert_equal(console_encode(UNICODE, 'UTF-8', force=True), b'hyv\xc3\xa4')

    def test_encoding_error(self):
        assert_equal(console_encode(UNICODE, 'ASCII'), 'hyv?')
        assert_equal(console_encode(UNICODE, 'ASCII', force=True), b'hyv?')

    def test_non_string(self):
        assert_equal(console_encode(42), '42')


if __name__ == '__main__':
    unittest.main()



import sys
from remoteserver import DirectResultRemoteServer


class Invalid:

    def non_dict_result_dict(self):
        return 42

    def invalid_result_dict(self):
        return {}

    def invalid_char_in_xml(self):
        return {'status': 'PASS', 'return': '\x00'}

    def exception(self, message):
        raise Exception(message)

    def shutdown(self):
        sys.exit()


if __name__ == '__main__':
    DirectResultRemoteServer(Invalid(), *sys.argv[1:])



import ast
import os
import unittest
import tempfile
from pathlib import Path

from robot.parsing import get_model, get_resource_model, ModelVisitor, ModelTransformer, Token
from robot.parsing.model.blocks import (
    File, For, If, ImplicitCommentSection, InvalidSection, Try, While,
    Keyword, KeywordSection, SettingSection, TestCase, TestCaseSection, VariableSection
)
from robot.parsing.model.statements import (
    Arguments, Break, Comment, Config, Continue, Documentation, ForHeader, End, ElseHeader,
    ElseIfHeader, EmptyLine, Error, IfHeader, InlineIfHeader, TryHeader, ExceptHeader,
    FinallyHeader, KeywordCall, KeywordName, Return, ReturnSetting, ReturnStatement,
    SectionHeader, TestCaseName, Variable, WhileHeader
)
from robot.utils.asserts import assert_equal, assert_raises_with_msg

from parsing_test_utils import assert_model, remove_non_data


DATA = '''\

*** Test Cases ***

Example
  # Comment
    Keyword    arg
    ...\targh

\t\t
*** Keywords ***
# Comment    continues
Keyword
    [Arguments]    ${arg1}    ${arg2}
    Log    Got ${arg1} and ${arg}!
    RETURN    x
'''
PATH = Path(os.getenv('TEMPDIR') or tempfile.gettempdir(), 'test_model.robot')
EXPECTED = File(sections=[
    ImplicitCommentSection(
        body=[
            EmptyLine([
                Token('EOL', '\n', 1, 0)
            ])
        ]
    ),
    TestCaseSection(
        header=SectionHeader([
            Token('TESTCASE HEADER', '*** Test Cases ***', 2, 0),
            Token('EOL', '\n', 2, 18)
        ]),
        body=[
            EmptyLine([Token('EOL', '\n', 3, 0)]),
            TestCase(
                header=TestCaseName([
                    Token('TESTCASE NAME', 'Example', 4, 0),
                    Token('EOL', '\n', 4, 7)
                ]),
                body=[
                    Comment([
                        Token('SEPARATOR', '  ', 5, 0),
                        Token('COMMENT', '# Comment', 5, 2),
                        Token('EOL', '\n', 5, 11),
                    ]),
                    KeywordCall([
                        Token('SEPARATOR', '    ', 6, 0),
                        Token('KEYWORD', 'Keyword', 6, 4),
                        Token('SEPARATOR', '    ', 6, 11),
                        Token('ARGUMENT', 'arg', 6, 15),
                        Token('EOL', '\n', 6, 18),
                        Token('SEPARATOR', '    ', 7, 0),
                        Token('CONTINUATION', '...', 7, 4),
                        Token('SEPARATOR', '\t', 7, 7),
                        Token('ARGUMENT', 'argh', 7, 8),
                        Token('EOL', '\n', 7, 12)
                    ]),
                    EmptyLine([Token('EOL', '\n', 8, 0)]),
                    EmptyLine([Token('EOL', '\t\t\n', 9, 0)])
                ]
            )
        ]
    ),
    KeywordSection(
        header=SectionHeader([
            Token('KEYWORD HEADER', '*** Keywords ***', 10, 0),
            Token('EOL', '\n', 10, 16)
        ]),
        body=[
            Comment([
                Token('COMMENT', '# Comment', 11, 0),
                Token('SEPARATOR', '    ', 11, 9),
                Token('COMMENT', 'continues', 11, 13),
                Token('EOL', '\n', 11, 22),
            ]),
            Keyword(
                header=KeywordName([
                    Token('KEYWORD NAME', 'Keyword', 12, 0),
                    Token('EOL', '\n', 12, 7)
                ]),
                body=[
                    Arguments([
                        Token('SEPARATOR', '    ', 13, 0),
                        Token('ARGUMENTS', '[Arguments]', 13, 4),
                        Token('SEPARATOR', '    ', 13, 15),
                        Token('ARGUMENT', '${arg1}', 13, 19),
                        Token('SEPARATOR', '    ', 13, 26),
                        Token('ARGUMENT', '${arg2}', 13, 30),
                        Token('EOL', '\n', 13, 37)
                    ]),
                    KeywordCall([
                        Token('SEPARATOR', '    ', 14, 0),
                        Token('KEYWORD', 'Log', 14, 4),
                        Token('SEPARATOR', '    ', 14, 7),
                        Token('ARGUMENT', 'Got ${arg1} and ${arg}!', 14, 11),
                        Token('EOL', '\n', 14, 34)
                    ]),
                    ReturnStatement([
                        Token('SEPARATOR', '    ', 15, 0),
                        Token('RETURN STATEMENT', 'RETURN', 15, 4),
                        Token('SEPARATOR', '    ', 15, 10),
                        Token('ARGUMENT', 'x', 15, 14),
                        Token('EOL', '\n', 15, 15)
                    ])
                ]
            )
        ]
    )
])


def get_and_assert_model(data, expected, depth=2):
    for data_only in True, False:
        model = get_model(data.strip(), data_only=data_only)
        if not data_only:
            remove_non_data(model)
        node = model.sections[0]
        for _ in range(depth):
            node = node.body[0]
        assert_model(node, expected)


class TestGetModel(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        PATH.write_text(DATA)

    @classmethod
    def tearDownClass(cls):
        PATH.unlink()

    def test_from_string(self):
        model = get_model(DATA)
        assert_model(model, EXPECTED)

    def test_from_path_as_string(self):
        model = get_model(str(PATH))
        assert_model(model, EXPECTED, source=PATH)

    def test_from_path_as_path(self):
        model = get_model(PATH)
        assert_model(model, EXPECTED, source=PATH)

    def test_from_open_file(self):
        with open(PATH) as f:
            model = get_model(f)
        assert_model(model, EXPECTED)


class TestSaveModel(unittest.TestCase):
    different_path = PATH.parent / 'different.robot'

    @classmethod
    def setUpClass(cls):
        PATH.write_text(DATA)

    @classmethod
    def tearDownClass(cls):
        PATH.unlink()
        if cls.different_path.exists:
            cls.different_path.unlink()

    def test_save_to_original_path(self):
        model = get_model(PATH)
        PATH.unlink()
        model.save()
        assert_model(get_model(PATH), EXPECTED, source=PATH)

    def test_save_to_different_path(self):
        model = get_model(PATH)
        path = self.different_path
        model.save(path)
        assert_model(get_model(path), EXPECTED, source=path)

    def test_save_to_original_path_as_str(self):
        model = get_model(str(PATH))
        PATH.unlink()
        model.save()
        assert_model(get_model(PATH), EXPECTED, source=PATH)

    def test_save_to_different_path_as_str(self):
        model = get_model(PATH)
        path = self.different_path
        model.save(Path(path))
        assert_model(get_model(path), EXPECTED, source=path)

    def test_save_to_original_fails_if_source_is_not_path(self):
        message = 'Saving model requires explicit output ' \
                  'when original source is not path.'
        assert_raises_with_msg(TypeError, message, get_model(DATA).save)
        with open(PATH) as f:
            assert_raises_with_msg(TypeError, message, get_model(f).save)


class TestForLoop(unittest.TestCase):

    def test_valid(self):
        data = '''
*** Test Cases ***
Example
    FOR    ${x}    IN    a    b    c
        Log    ${x}
    END
'''
        expected = For(
            header=ForHeader([
                Token(Token.FOR, 'FOR', 3, 4),
                Token(Token.VARIABLE, '${x}', 3, 11),
                Token(Token.FOR_SEPARATOR, 'IN', 3, 19),
                Token(Token.ARGUMENT, 'a', 3, 25),
                Token(Token.ARGUMENT, 'b', 3, 30),
                Token(Token.ARGUMENT, 'c', 3, 35),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Log', 4, 8),
                             Token(Token.ARGUMENT, '${x}', 4, 15)])
            ],
            end=End([
                Token(Token.END, 'END', 5, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_enumerate_with_start(self):
        data = '''
*** Test Cases ***
Example
    FOR    ${x}    IN ENUMERATE    @{stuff}    start=1
        Log    ${x}
    END
'''
        expected = For(
            header=ForHeader([
                Token(Token.FOR, 'FOR', 3, 4),
                Token(Token.VARIABLE, '${x}', 3, 11),
                Token(Token.FOR_SEPARATOR, 'IN ENUMERATE', 3, 19),
                Token(Token.ARGUMENT, '@{stuff}', 3, 35),
                Token(Token.OPTION, 'start=1', 3, 47),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Log', 4, 8),
                             Token(Token.ARGUMENT, '${x}', 4, 15)])
            ],
            end=End([
                Token(Token.END, 'END', 5, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_nested(self):
        data = '''
*** Test Cases ***
Example
    FOR    ${x}    IN    1    start=has no special meaning here
        FOR    ${y}    IN RANGE    ${x}
            Log    ${y}
        END
    END
'''
        expected = For(
            header=ForHeader([
                Token(Token.FOR, 'FOR', 3, 4),
                Token(Token.VARIABLE, '${x}', 3, 11),
                Token(Token.FOR_SEPARATOR, 'IN', 3, 19),
                Token(Token.ARGUMENT, '1', 3, 25),
                Token(Token.ARGUMENT, 'start=has no special meaning here', 3, 30),
            ]),
            body=[
                For(
                    header=ForHeader([
                        Token(Token.FOR, 'FOR', 4, 8),
                        Token(Token.VARIABLE, '${y}', 4, 15),
                        Token(Token.FOR_SEPARATOR, 'IN RANGE', 4, 23),
                        Token(Token.ARGUMENT, '${x}', 4, 35),
                    ]),
                    body=[
                        KeywordCall([Token(Token.KEYWORD, 'Log', 5, 12),
                                     Token(Token.ARGUMENT, '${y}', 5, 19)])
                    ],
                    end=End([
                        Token(Token.END, 'END', 6, 8)
                    ])
                )
            ],
            end=End([
                Token(Token.END, 'END', 7, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_invalid(self):
        data1 = '''
*** Test Cases ***
Example
    FOR

    END    ooops
'''
        data2 = '''
*** Test Cases ***
Example
    FOR    wrong    IN
'''
        expected1 = For(
            header=ForHeader(
                tokens=[Token(Token.FOR, 'FOR', 3, 4)],
                errors=('FOR loop has no loop variables.',
                        "FOR loop has no 'IN' or other valid separator."),
            ),
            end=End(
                tokens=[Token(Token.END, 'END', 5, 4),
                        Token(Token.ARGUMENT, 'ooops', 5, 11)],
                errors=("END does not accept arguments, got 'ooops'.",)
            ),
            errors=('FOR loop cannot be empty.',)
        )
        expected2 = For(
            header=ForHeader(
                tokens=[Token(Token.FOR, 'FOR', 3, 4),
                        Token(Token.VARIABLE, 'wrong', 3, 11),
                        Token(Token.FOR_SEPARATOR, 'IN', 3, 20)],
                errors=("FOR loop has invalid loop variable 'wrong'.",
                        "FOR loop has no loop values."),
            ),
            errors=('FOR loop cannot be empty.',
                    'FOR loop must have closing END.')
        )
        get_and_assert_model(data1, expected1)
        get_and_assert_model(data2, expected2)


class TestWhileLoop(unittest.TestCase):

    def test_valid(self):
        data = '''
*** Test Cases ***
Example
    WHILE    True
        Log    ${x}
    END
'''
        expected = While(
            header=WhileHeader([
                Token(Token.WHILE, 'WHILE', 3, 4),
                Token(Token.ARGUMENT, 'True', 3, 13),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Log', 4, 8),
                             Token(Token.ARGUMENT, '${x}', 4, 15)])
            ],
            end=End([
                Token(Token.END, 'END', 5, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_limit(self):
        data = '''
*** Test Cases ***
Example
    WHILE    True    limit=100
        Log    ${x}
    END
'''
        expected = While(
            header=WhileHeader([
                Token(Token.WHILE, 'WHILE', 3, 4),
                Token(Token.ARGUMENT, 'True', 3, 13),
                Token(Token.OPTION, 'limit=100', 3, 21),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Log', 4, 8),
                             Token(Token.ARGUMENT, '${x}', 4, 15)])
            ],
            end=End([
                Token(Token.END, 'END', 5, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_on_limit_message(self):
        data = '''
*** Test Cases ***
Example
    WHILE    True    limit=10s    on_limit=pass    on_limit_message=Error message
        Log    ${x}
    END
'''
        expected = While(
            header=WhileHeader([
                Token(Token.WHILE, 'WHILE', 3, 4),
                Token(Token.ARGUMENT, 'True', 3, 13),
                Token(Token.OPTION, 'limit=10s', 3, 21),
                Token(Token.OPTION, 'on_limit=pass', 3, 34),
                Token(Token.OPTION, 'on_limit_message=Error message', 3, 51)
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Log', 4, 8),
                             Token(Token.ARGUMENT, '${x}', 4, 15)])
            ],
            end=End([
                Token(Token.END, 'END', 5, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_invalid(self):
        data = '''
*** Test Cases ***
Example
    WHILE    too    many    values    !
        # Empty body
    END
'''
        expected = While(
            header=WhileHeader(
                tokens=[Token(Token.WHILE, 'WHILE', 3, 4),
                        Token(Token.ARGUMENT, 'too', 3, 13),
                        Token(Token.ARGUMENT, 'many', 3, 20),
                        Token(Token.ARGUMENT, 'values', 3, 28),
                        Token(Token.ARGUMENT, '!', 3, 38)],
                errors=("WHILE loop cannot have more than one condition, "
                        "got 'too', 'many', 'values' and '!'.",)
            ),
            end=End([
                Token(Token.END, 'END', 5, 4)
            ]),
            errors=('WHILE loop cannot be empty.',)
        )
        get_and_assert_model(data, expected)


class TestIf(unittest.TestCase):

    def test_if(self):
        data = '''
*** Test Cases ***
Example
    IF    True
        Keyword
        Another    argument
    END
    '''
        expected = If(
            header=IfHeader([
                Token(Token.IF, 'IF', 3, 4),
                Token(Token.ARGUMENT, 'True', 3, 10),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Keyword', 4, 8)]),
                KeywordCall([Token(Token.KEYWORD, 'Another', 5, 8),
                             Token(Token.ARGUMENT, 'argument', 5, 19)])
            ],
            end=End([Token(Token.END, 'END', 6, 4)])
        )
        get_and_assert_model(data, expected)

    def test_if_else_if_else(self):
        data = '''
*** Test Cases ***
Example
    IF    True
        K1
    ELSE IF    False
        K2
    ELSE
        K3
    END
    '''
        expected = If(
            header=IfHeader([
                Token(Token.IF, 'IF', 3, 4),
                Token(Token.ARGUMENT, 'True', 3, 10),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'K1', 4, 8)])
            ],
            orelse=If(
                header=ElseIfHeader([
                    Token(Token.ELSE_IF, 'ELSE IF', 5, 4),
                    Token(Token.ARGUMENT, 'False', 5, 15),
                ]),
                body=[
                    KeywordCall([Token(Token.KEYWORD, 'K2', 6, 8)])
                ],
                orelse=If(
                    header=ElseHeader([
                        Token(Token.ELSE, 'ELSE', 7, 4),
                    ]),
                    body=[
                        KeywordCall([Token(Token.KEYWORD, 'K3', 8, 8)])
                    ],
                )
            ),
            end=End([Token(Token.END, 'END', 9, 4)])
        )
        get_and_assert_model(data, expected)

    def test_nested(self):
        data = '''
*** Test Cases ***
Example
    IF    ${x}
        Log    ${x}
        IF    ${y}
            Log    ${y}
        ELSE
            Log    ${z}
        END
    END
'''
        expected = If(
            header=IfHeader([
                Token(Token.IF, 'IF', 3, 4),
                Token(Token.ARGUMENT, '${x}', 3, 10),
            ]),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Log', 4, 8),
                             Token(Token.ARGUMENT, '${x}', 4, 15)]),
                If(
                    header=IfHeader([
                        Token(Token.IF, 'IF', 5, 8),
                        Token(Token.ARGUMENT, '${y}', 5, 14),
                    ]),
                    body=[
                        KeywordCall([Token(Token.KEYWORD, 'Log', 6, 12),
                                     Token(Token.ARGUMENT, '${y}', 6, 19)])
                    ],
                    orelse=If(
                        header=ElseHeader([
                            Token(Token.ELSE, 'ELSE', 7, 8)
                        ]),
                        body=[
                            KeywordCall([Token(Token.KEYWORD, 'Log', 8, 12),
                                         Token(Token.ARGUMENT, '${z}', 8, 19)])
                        ]
                    ),
                    end=End([
                        Token(Token.END, 'END', 9, 8)
                    ])
                )
            ],
            end=End([
                Token(Token.END, 'END', 10, 4)
            ])
        )
        get_and_assert_model(data, expected)

    def test_invalid(self):
        data1 = '''
*** Test Cases ***
Example
    IF
    ELSE    ooops
        # Empty
    ELSE IF

    END    ooops
'''
        data2 = '''
*** Test Cases ***
Example
    IF
'''
        expected1 = If(
            header=IfHeader(
                tokens=[Token(Token.IF, 'IF', 3, 4)],
                errors=('IF must have a condition.',)
            ),
            orelse=If(
                header=ElseHeader(
                    tokens=[Token(Token.ELSE, 'ELSE', 4, 4),
                            Token(Token.ARGUMENT, 'ooops', 4, 12)],
                    errors=("ELSE does not accept arguments, got 'ooops'.",)
                ),
                orelse=If(
                    header=ElseIfHeader(
                        tokens=[Token(Token.ELSE_IF, 'ELSE IF', 6, 4)],
                        errors=('ELSE IF must have a condition.',)
                    ),
                    errors=('ELSE IF branch cannot be empty.',)
                ),
                errors=('ELSE branch cannot be empty.',)
            ),
            end=End(
                tokens=[Token(Token.END, 'END', 8, 4),
                        Token(Token.ARGUMENT, 'ooops', 8, 11)],
                errors=("END does not accept arguments, got 'ooops'.",)
            ),
            errors=('IF branch cannot be empty.',
                    'ELSE IF not allowed after ELSE.')
        )
        expected2 = If(
            header=IfHeader(
                tokens=[Token(Token.IF, 'IF', 3, 4)],
                errors=('IF must have a condition.',)
            ),
            errors=('IF branch cannot be empty.',
                    'IF must have closing END.')
        )
        get_and_assert_model(data1, expected1)
        get_and_assert_model(data2, expected2)


class TestInlineIf(unittest.TestCase):

    def test_if(self):
        data = '''
*** Test Cases ***
Example
    IF    True    Keyword
'''
        expected = If(
            header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 4),
                                   Token(Token.ARGUMENT, 'True', 3, 10)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'Keyword', 3, 18)])],
            end=End([Token(Token.END, '', 3, 25)])
        )
        get_and_assert_model(data, expected)

    def test_if_else_if_else(self):
        data = '''
*** Test Cases ***
Example
    IF    True    K1    ELSE IF    False    K2    ELSE    K3
'''
        expected = If(
            header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 4),
                                   Token(Token.ARGUMENT, 'True', 3, 10)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'K1', 3, 18)])],
            orelse=If(
                header=ElseIfHeader([Token(Token.ELSE_IF, 'ELSE IF', 3, 24),
                                     Token(Token.ARGUMENT, 'False', 3, 35)]),
                body=[KeywordCall([Token(Token.KEYWORD, 'K2', 3, 44)])],
                orelse=If(
                    header=ElseHeader([Token(Token.ELSE, 'ELSE', 3, 50)]),
                    body=[KeywordCall([Token(Token.KEYWORD, 'K3', 3, 58)])],
                )
            ),
            end=End([Token(Token.END, '', 3, 60)])
        )
        get_and_assert_model(data, expected)

    def test_nested(self):
        data = '''
*** Test Cases ***
Example
    IF    ${x}    IF    ${y}    K1    ELSE    IF    ${z}    K2
'''
        expected = If(
            header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 4),
                                   Token(Token.ARGUMENT, '${x}', 3, 10)]),
            body=[If(
                header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 18),
                                       Token(Token.ARGUMENT, '${y}', 3, 24)]),
                body=[KeywordCall([Token(Token.KEYWORD, 'K1', 3, 32)])],
                orelse=If(
                    header=ElseHeader([Token(Token.ELSE, 'ELSE', 3, 38)]),
                    body=[If(
                        header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 46),
                                               Token(Token.ARGUMENT, '${z}', 3, 52)]),
                        body=[KeywordCall([Token(Token.KEYWORD, 'K2', 3, 60)])],
                        end=End([Token(Token.END, '', 3, 62)]),
                    )],
                ),
                errors=('Inline IF cannot be nested.',),
            )],
            errors=('Inline IF cannot be nested.',),
        )
        get_and_assert_model(data, expected)

    def test_assign(self):
        data = '''
*** Test Cases ***
Example
    ${x} =    IF    True    K1    ELSE    K2
'''
        expected = If(
            header=InlineIfHeader([Token(Token.ASSIGN, '${x} =', 3, 4),
                                   Token(Token.INLINE_IF, 'IF', 3, 14),
                                   Token(Token.ARGUMENT, 'True', 3, 20)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'K1', 3, 28)])],
            orelse=If(
                header=ElseHeader([Token(Token.ELSE, 'ELSE', 3, 34)]),
                body=[KeywordCall([Token(Token.KEYWORD, 'K2', 3, 42)])],
            ),
            end=End([Token(Token.END, '', 3, 44)])
        )
        get_and_assert_model(data, expected)

    def test_assign_only_inside(self):
        data = '''
*** Test Cases ***
Example
    IF    ${cond}    ${assign}
'''
        expected = If(
            header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 4),
                                   Token(Token.ARGUMENT, '${cond}', 3, 10)]),
            body=[KeywordCall([Token(Token.ASSIGN, '${assign}', 3, 21)])],
            end=End([Token(Token.END, '', 3, 30)]),
            errors=('Inline IF branches cannot contain assignments.',)
        )
        get_and_assert_model(data, expected)

    def test_invalid(self):
        data1 = '''
*** Test Cases ***
Example
    ${x} =    ${y}    IF    ELSE    ooops    ELSE IF
'''
        data2 = '''
*** Test Cases ***
Example
    IF    e    K    ELSE
'''
        expected1 = If(
            header=InlineIfHeader([Token(Token.ASSIGN, '${x} =', 3, 4),
                                   Token(Token.ASSIGN, '${y}', 3, 14),
                                   Token(Token.INLINE_IF, 'IF', 3, 22),
                                   Token(Token.ARGUMENT, 'ELSE', 3, 28)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'ooops', 3, 36)])],
            orelse=If(
                header=ElseIfHeader([Token(Token.ELSE_IF, 'ELSE IF', 3, 45)],
                                    errors=('ELSE IF must have a condition.',)),
                errors=('ELSE IF branch cannot be empty.',),
            ),
            end=End([Token(Token.END, '', 3, 52)])
        )
        expected2 = If(
            header=InlineIfHeader([Token(Token.INLINE_IF, 'IF', 3, 4),
                                   Token(Token.ARGUMENT, 'e', 3, 10)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'K', 3, 15)])],
            orelse=If(
                header=ElseHeader([Token(Token.ELSE, 'ELSE', 3, 20)]),
                errors=('ELSE branch cannot be empty.',),
            ),
            end=End([Token(Token.END, '', 3, 24)])
        )
        get_and_assert_model(data1, expected1)
        get_and_assert_model(data2, expected2)


class TestTry(unittest.TestCase):

    def test_try_except_else_finally(self):
        data = '''
*** Test Cases ***
Example
    TRY
        Fail    Oh no!
    EXCEPT    does not match
        No operation
    EXCEPT    AS    ${exp}
        Log    Catch
    ELSE
        No operation
    FINALLY
        Log    finally here!
    END
'''
        expected = Try(
            header=TryHeader([Token(Token.TRY, 'TRY', 3, 4)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'Fail', 4, 8),
                               Token(Token.ARGUMENT, 'Oh no!', 4, 16)])],
            next=Try(
                header=ExceptHeader([Token(Token.EXCEPT, 'EXCEPT', 5, 4),
                                     Token(Token.ARGUMENT, 'does not match', 5, 14)]),
                body=[KeywordCall((Token(Token.KEYWORD, 'No operation', 6, 8),))],
                next=Try(
                    header=ExceptHeader((Token(Token.EXCEPT, 'EXCEPT', 7, 4),
                                         Token(Token.AS, 'AS', 7, 14),
                                         Token(Token.VARIABLE, '${exp}', 7, 20))),
                    body=[KeywordCall((Token(Token.KEYWORD, 'Log', 8, 8),
                                       Token(Token.ARGUMENT, 'Catch', 8, 15)))],
                    next=Try(
                        header=ElseHeader((Token(Token.ELSE, 'ELSE', 9, 4),)),
                        body=[KeywordCall((Token(Token.KEYWORD, 'No operation', 10, 8),))],
                        next=Try(
                            header=FinallyHeader((Token(Token.FINALLY, 'FINALLY', 11, 4),)),
                            body=[KeywordCall((Token(Token.KEYWORD, 'Log', 12, 8),
                                               Token(Token.ARGUMENT, 'finally here!', 12, 15)))]
                        )
                    )
                )
            ),
            end=End([Token(Token.END, 'END', 13, 4)])
        )
        get_and_assert_model(data, expected)

    def test_invalid(self):
        data = '''
*** Test Cases ***
Example
    TRY             invalid
    ELSE            invalid

    FINALLY         invalid
    #
    EXCEPT    AS    invalid
'''
        expected = Try(
            header=TryHeader(
                tokens=[Token(Token.TRY, 'TRY', 3, 4),
                        Token(Token.ARGUMENT, 'invalid', 3, 20)],
                errors=("TRY does not accept arguments, got 'invalid'.",)
            ),
            next=Try(
                header=ElseHeader(
                    tokens=[Token(Token.ELSE, 'ELSE', 4, 4),
                            Token(Token.ARGUMENT, 'invalid', 4, 20)],
                    errors=("ELSE does not accept arguments, got 'invalid'.",)
                ),
                errors=('ELSE branch cannot be empty.',),
                next=Try(
                    header=FinallyHeader(
                        tokens=[Token(Token.FINALLY, 'FINALLY', 6, 4),
                                Token(Token.ARGUMENT, 'invalid', 6, 20)],
                        errors=("FINALLY does not accept arguments, got 'invalid'.",)
                    ),
                    errors=('FINALLY branch cannot be empty.',),
                    next=Try(
                        header=ExceptHeader(
                            tokens=[Token(Token.EXCEPT, 'EXCEPT', 8, 4),
                                    Token(Token.AS, 'AS', 8, 14),
                                    Token(Token.VARIABLE, 'invalid', 8, 20)],
                            errors=("EXCEPT's AS variable 'invalid' is invalid.",)
                        ),
                        errors=('EXCEPT branch cannot be empty.',)
                    )
                ),
            ),
            errors=('TRY branch cannot be empty.',
                    'EXCEPT not allowed after ELSE.',
                    'EXCEPT not allowed after FINALLY.',
                    'TRY must have closing END.')
        )
        get_and_assert_model(data, expected)


class TestVariables(unittest.TestCase):

    def test_valid(self):
        data = '''
*** Variables ***
${x}      value
@{y}=     two    values
&{z} =    one=item
${x${y}}  nested name
'''
        expected = VariableSection(
            header=SectionHeader(
                tokens=[Token(Token.VARIABLE_HEADER, '*** Variables ***', 1, 0)]
            ),
            body=[
                Variable([Token(Token.VARIABLE, '${x}', 2, 0),
                         Token(Token.ARGUMENT, 'value', 2, 10)]),
                Variable([Token(Token.VARIABLE, '@{y}=', 3, 0),
                          Token(Token.ARGUMENT, 'two', 3, 10),
                          Token(Token.ARGUMENT, 'values', 3, 17)]),
                Variable([Token(Token.VARIABLE, '&{z} =', 4, 0),
                          Token(Token.ARGUMENT, 'one=item', 4, 10)]),
                Variable([Token(Token.VARIABLE, '${x${y}}', 5, 0),
                          Token(Token.ARGUMENT, 'nested name', 5, 10)]),
            ]
        )
        get_and_assert_model(data, expected, depth=0)

    def test_invalid(self):
        data = '''
*** Variables ***
Ooops     I did it again
${}       invalid
${x}==    invalid
${not     closed
          invalid
&{dict}   invalid    ${invalid}
'''
        expected = VariableSection(
            header=SectionHeader(
                tokens=[Token(Token.VARIABLE_HEADER, '*** Variables ***', 1, 0)]
            ),
            body=[
                Variable(
                    tokens=[Token(Token.VARIABLE, 'Ooops', 2, 0),
                            Token(Token.ARGUMENT, 'I did it again', 2, 10)],
                    errors=("Invalid variable name 'Ooops'.",)
                ),
                Variable(
                    tokens=[Token(Token.VARIABLE, '${}', 3, 0),
                            Token(Token.ARGUMENT, 'invalid', 3, 10)],
                    errors=("Invalid variable name '${}'.",)
                ),
                Variable(
                    tokens=[Token(Token.VARIABLE, '${x}==', 4, 0),
                            Token(Token.ARGUMENT, 'invalid', 4, 10)],
                    errors=("Invalid variable name '${x}=='.",)
                ),
                Variable(
                    tokens=[Token(Token.VARIABLE, '${not', 5, 0),
                            Token(Token.ARGUMENT, 'closed', 5, 10)],
                    errors=("Invalid variable name '${not'.",)
                ),
                Variable(
                    tokens=[Token(Token.VARIABLE, '', 6, 0),
                            Token(Token.ARGUMENT, 'invalid', 6, 10)],
                    errors=("Invalid variable name ''.",)
                ),
                Variable(
                    tokens=[Token(Token.VARIABLE, '&{dict}', 7, 0),
                            Token(Token.ARGUMENT, 'invalid', 7, 10),
                            Token(Token.ARGUMENT, '${invalid}', 7, 21)],
                    errors=("Invalid dictionary variable item 'invalid'. "
                            "Items must use 'name=value' syntax or be dictionary variables themselves.",
                            "Invalid dictionary variable item '${invalid}'. "
                            "Items must use 'name=value' syntax or be dictionary variables themselves.")
                ),
            ]
        )
        get_and_assert_model(data, expected, depth=0)


class TestTestCase(unittest.TestCase):

    def test_empty_test(self):
        data = '''
*** Test Cases ***
Empty
    [Documentation]    Settings aren't enough.
'''
        expected = TestCase(
            header=TestCaseName(
                tokens=[Token(Token.TESTCASE_NAME, 'Empty', 2, 0)]
            ),
            body=[
                Documentation(
                    tokens=[Token(Token.DOCUMENTATION, '[Documentation]', 3, 4),
                            Token(Token.ARGUMENT, "Settings aren't enough.", 3, 23)]
                ),
            ],
            errors=('Test cannot be empty.',)
        )
        get_and_assert_model(data, expected, depth=1)

    def test_empty_test_name(self):
        data = '''
*** Test Cases ***
    Keyword
'''
        expected = TestCase(
            header=TestCaseName(
                tokens=[Token(Token.TESTCASE_NAME, '', 2, 0)],
                errors=('Test name cannot be empty.',)
            ),
            body=[KeywordCall(tokens=[Token(Token.KEYWORD, 'Keyword', 2, 4)])]
        )
        get_and_assert_model(data, expected, depth=1)

    def test_invalid_task(self):
        data = '''
*** Tasks ***
    [Documentation]    Empty name and body.
'''
        expected = TestCase(
            header=TestCaseName(
                tokens=[Token(Token.TESTCASE_NAME, '', 2, 0)],
                errors=('Task name cannot be empty.',)
            ),
            body=[
                Documentation(
                    tokens=[Token(Token.DOCUMENTATION, '[Documentation]', 2, 4),
                            Token(Token.ARGUMENT, 'Empty name and body.', 2, 23)]
                ),
            ],
            errors=('Task cannot be empty.',)
        )
        get_and_assert_model(data, expected, depth=1)


class TestUserKeyword(unittest.TestCase):

    def test_invalid_arg_spec(self):
        data = '''
*** Keywords ***
Invalid
    [Arguments]    ooops    ${optional}=default    ${required}
    ...    @{too}    @{many}    &{notlast}    ${x}
    Keyword
'''
        expected = Keyword(
            header=KeywordName(
                tokens=[Token(Token.KEYWORD_NAME, 'Invalid', 2, 0)]
            ),
            body=[
                Arguments(
                    tokens=[Token(Token.ARGUMENTS, '[Arguments]', 3, 4),
                            Token(Token.ARGUMENT, 'ooops', 3, 19),
                            Token(Token.ARGUMENT, '${optional}=default', 3, 28),
                            Token(Token.ARGUMENT, '${required}', 3, 51),
                            Token(Token.ARGUMENT, '@{too}', 4, 11),
                            Token(Token.ARGUMENT, '@{many}', 4, 21),
                            Token(Token.ARGUMENT, '&{notlast}', 4, 32),
                            Token(Token.ARGUMENT, '${x}', 4, 46)],
                    errors=("Invalid argument syntax 'ooops'.",
                            'Non-default argument after default arguments.',
                            'Cannot have multiple varargs.',
                            'Only last argument can be kwargs.')
                ),
                KeywordCall(
                    tokens=[Token(Token.KEYWORD, 'Keyword', 5, 4)])
            ],
        )
        get_and_assert_model(data, expected, depth=1)

    def test_empty(self):
        data = '''
*** Keywords ***
Empty
    [Arguments]    ${ok}
'''
        expected = Keyword(
            header=KeywordName(
                tokens=[Token(Token.KEYWORD_NAME, 'Empty', 2, 0)]
            ),
            body=[
                Arguments(
                    tokens=[Token(Token.ARGUMENTS, '[Arguments]', 3, 4),
                            Token(Token.ARGUMENT, '${ok}', 3, 19)]
                ),
            ],
            errors=('User keyword cannot be empty.',)
        )
        get_and_assert_model(data, expected, depth=1)

    def test_empty_name(self):
        data = '''
*** Keywords ***
    Keyword
'''
        expected = Keyword(
            header=KeywordName(
                tokens=[Token(Token.KEYWORD_NAME, '', 2, 0)],
                errors=('User keyword name cannot be empty.',)
            ),
            body=[KeywordCall(tokens=[Token(Token.KEYWORD, 'Keyword', 2, 4)])]
        )
        get_and_assert_model(data, expected, depth=1)


class TestControlStatements(unittest.TestCase):

    def test_return(self):
        data = '''
*** Keywords ***
Name
    Return    RETURN
    RETURN    RETURN
'''
        expected = Keyword(
            header=KeywordName(
                tokens=[Token(Token.KEYWORD_NAME, 'Name', 2, 0)]
            ),
            body=[
                KeywordCall([Token(Token.KEYWORD, 'Return', 3, 4),
                             Token(Token.ARGUMENT, 'RETURN', 3, 14)]),
                ReturnStatement([Token(Token.RETURN_STATEMENT, 'RETURN', 4, 4),
                                 Token(Token.ARGUMENT, 'RETURN', 4, 14)])
            ],
        )
        get_and_assert_model(data, expected, depth=1)

    def test_break(self):
        data = '''
*** Keywords ***
Name
    WHILE    True
        Break    BREAK
        BREAK
    END
'''
        expected = While(
            header=WhileHeader([Token(Token.WHILE, 'WHILE', 3, 4),
                                Token(Token.ARGUMENT, 'True', 3, 13)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'Break', 4, 8),
                               Token(Token.ARGUMENT, 'BREAK', 4, 17)]),
                  Break([Token(Token.BREAK, 'BREAK', 5, 8)])],
            end=End([Token(Token.END, 'END', 6, 4)])
        )
        get_and_assert_model(data, expected)

    def test_continue(self):
        data = '''
*** Keywords ***
Name
    FOR    ${x}    IN    @{stuff}
        Continue    CONTINUE
        CONTINUE
    END
'''
        expected = For(
            header=ForHeader([Token(Token.FOR, 'FOR', 3, 4),
                              Token(Token.VARIABLE, '${x}', 3, 11),
                              Token(Token.FOR_SEPARATOR, 'IN', 3, 19),
                              Token(Token.ARGUMENT, '@{stuff}', 3, 25)]),
            body=[KeywordCall([Token(Token.KEYWORD, 'Continue', 4, 8),
                               Token(Token.ARGUMENT, 'CONTINUE', 4, 20)]),
                  Continue([Token(Token.CONTINUE, 'CONTINUE', 5, 8)])],
            end=End([Token(Token.END, 'END', 6, 4)])
        )
        get_and_assert_model(data, expected)


class TestDocumentation(unittest.TestCase):

    def test_empty(self):
        data = '''\
*** Settings ***
Documentation
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.EOL, '\n', 2, 13)]
        )
        self._verify_documentation(data, expected, '')

    def test_one_line(self):
        data = '''\
*** Settings ***
Documentation    Hello!
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'Hello!', 2, 17),
                    Token(Token.EOL, '\n', 2, 23)]
        )
        self._verify_documentation(data, expected, 'Hello!')

    def test_multi_part(self):
        data = '''\
*** Settings ***
Documentation    Hello    world
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'Hello', 2, 17),
                    Token(Token.SEPARATOR, '    ', 2, 22),
                    Token(Token.ARGUMENT, 'world', 2, 26),
                    Token(Token.EOL, '\n', 2, 31)]
        )
        self._verify_documentation(data, expected, 'Hello    world')

    def test_multi_line(self):
        data = '''\
*** Settings ***
Documentation    Documentation
...              in
...              multiple lines
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'Documentation', 2, 17),
                    Token(Token.EOL, '\n', 2, 30),
                    Token(Token.CONTINUATION, '...', 3, 0),
                    Token(Token.SEPARATOR, '              ', 3, 3),
                    Token(Token.ARGUMENT, 'in', 3, 17),
                    Token(Token.EOL, '\n', 3, 19),
                    Token(Token.CONTINUATION, '...', 4, 0),
                    Token(Token.SEPARATOR, '              ', 4, 3),
                    Token(Token.ARGUMENT, 'multiple lines', 4, 17),
                    Token(Token.EOL, '\n', 4, 31)]
        )
        self._verify_documentation(data, expected, 'Documentation\nin\nmultiple lines')

    def test_multi_line_with_empty_lines(self):
        data = '''\
*** Settings ***
Documentation    Documentation
...
...              with empty
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'Documentation', 2, 17),
                    Token(Token.EOL, '\n', 2, 30),
                    Token(Token.CONTINUATION, '...', 3, 0),
                    Token(Token.ARGUMENT, '', 3, 3),
                    Token(Token.EOL, '\n', 3, 3),
                    Token(Token.CONTINUATION, '...', 4, 0),
                    Token(Token.SEPARATOR, '              ', 4, 3),
                    Token(Token.ARGUMENT, 'with empty', 4, 17),
                    Token(Token.EOL, '\n', 4, 27)]
        )
        self._verify_documentation(data, expected, 'Documentation\n\nwith empty')

    def test_no_automatic_newline_after_literal_newline(self):
        data = '''\
*** Settings ***
Documentation    No automatic\\n
...              newline
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'No automatic\\n', 2, 17),
                    Token(Token.EOL, '\n', 2, 31),
                    Token(Token.CONTINUATION, '...', 3, 0),
                    Token(Token.SEPARATOR, '              ', 3, 3),
                    Token(Token.ARGUMENT, 'newline', 3, 17),
                    Token(Token.EOL, '\n', 3, 24)]
        )
        self._verify_documentation(data, expected, 'No automatic\\nnewline')

    def test_no_automatic_newline_after_backlash(self):
        data = '''\
*** Settings ***
Documentation    No automatic \\
...              newline\\\\\\
...              and remove\\    trailing\\\\    back\\slashes\\\\\\
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'No automatic \\', 2, 17),
                    Token(Token.EOL, '\n', 2, 31),
                    Token(Token.CONTINUATION, '...', 3, 0),
                    Token(Token.SEPARATOR, '              ', 3, 3),
                    Token(Token.ARGUMENT, 'newline\\\\\\', 3, 17),
                    Token(Token.EOL, '\n', 3, 27),
                    Token(Token.CONTINUATION, '...', 4, 0),
                    Token(Token.SEPARATOR, '              ', 4, 3),
                    Token(Token.ARGUMENT, 'and remove\\', 4, 17),
                    Token(Token.SEPARATOR, '    ', 4, 28),
                    Token(Token.ARGUMENT, 'trailing\\\\', 4, 32),
                    Token(Token.SEPARATOR, '    ', 4, 42),
                    Token(Token.ARGUMENT, 'back\\slashes\\\\\\', 4, 46),
                    Token(Token.EOL, '\n', 4, 61)]
        )
        self._verify_documentation(data, expected,
                                   'No automatic newline\\\\'
                                   'and remove    trailing\\\\    back\\slashes\\\\')

    def test_preserve_indentation(self):
        data = '''\
*** Settings ***
Documentation
...    Example:
...
...        - list with
...        - two
...          items
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.EOL, '\n', 2, 13),
                    Token(Token.CONTINUATION, '...', 3, 0),
                    Token(Token.SEPARATOR, '    ', 3, 3),
                    Token(Token.ARGUMENT, 'Example:', 3, 7),
                    Token(Token.EOL, '\n', 3, 15),
                    Token(Token.CONTINUATION, '...', 4, 0),
                    Token(Token.ARGUMENT, '', 4, 3),
                    Token(Token.EOL, '\n', 4, 3),
                    Token(Token.CONTINUATION, '...', 5, 0),
                    Token(Token.SEPARATOR, '        ', 5, 3),
                    Token(Token.ARGUMENT, '- list with', 5, 11),
                    Token(Token.EOL, '\n', 5, 22),
                    Token(Token.CONTINUATION, '...', 6, 0),
                    Token(Token.SEPARATOR, '        ', 6, 3),
                    Token(Token.ARGUMENT, '- two', 6, 11),
                    Token(Token.EOL, '\n', 6, 16),
                    Token(Token.CONTINUATION, '...', 7, 0),
                    Token(Token.SEPARATOR, '          ', 7, 3),
                    Token(Token.ARGUMENT, 'items', 7, 13),
                    Token(Token.EOL, '\n', 7, 18)]
        )
        self._verify_documentation(data, expected, '''\
Example:

    - list with
    - two
      items''')

    def test_preserve_indentation_with_data_on_first_doc_row(self):
        data = '''\
*** Settings ***
Documentation    Example:
...
...      - list with
...      - two
...        items
'''
        expected = Documentation(
            tokens=[Token(Token.DOCUMENTATION, 'Documentation', 2, 0),
                    Token(Token.SEPARATOR, '    ', 2, 13),
                    Token(Token.ARGUMENT, 'Example:', 2, 17),
                    Token(Token.EOL, '\n', 2, 25),
                    Token(Token.CONTINUATION, '...', 3, 0),
                    Token(Token.ARGUMENT, '', 3, 3),
                    Token(Token.EOL, '\n', 3, 3),
                    Token(Token.CONTINUATION, '...', 4, 0),
                    Token(Token.SEPARATOR, '      ', 4, 3),
                    Token(Token.ARGUMENT, '- list with', 4, 9),
                    Token(Token.EOL, '\n', 4, 20),
                    Token(Token.CONTINUATION, '...', 5, 0),
                    Token(Token.SEPARATOR, '      ', 5, 3),
                    Token(Token.ARGUMENT, '- two', 5, 9),
                    Token(Token.EOL, '\n', 5, 14),
                    Token(Token.CONTINUATION, '...', 6, 0),
                    Token(Token.SEPARATOR, '        ', 6, 3),
                    Token(Token.ARGUMENT, 'items', 6, 11),
                    Token(Token.EOL, '\n', 6, 16)]
        )
        self._verify_documentation(data, expected, '''\
Example:

- list with
- two
  items''')

    def _verify_documentation(self, data, expected, value):
        # Model has both EOLs and line numbers.
        doc = get_model(data).sections[0].body[0]
        assert_model(doc, expected)
        assert_equal(doc.value, value)
        # Model has only line numbers, no EOLs or other non-data tokens.
        doc = get_model(data, data_only=True).sections[0].body[0]
        expected.tokens = [token for token in expected.tokens
                           if token.type not in Token.NON_DATA_TOKENS]
        assert_model(doc, expected)
        assert_equal(doc.value, value)
        # Model has only EOLS, no line numbers.
        doc = Documentation.from_params(value)
        assert_equal(doc.value, value)
        # Model has no EOLs nor line numbers. Everything is just one line.
        doc.tokens = [token for token in doc.tokens if token.type != Token.EOL]
        assert_equal(doc.value, ' '.join(value.splitlines()))


class TestError(unittest.TestCase):

    def test_get_errors_from_tokens(self):
        assert_equal(Error([Token('ERROR', error='xxx')]).errors,
                     ('xxx',))
        assert_equal(Error([Token('ERROR', error='xxx'),
                            Token('ARGUMENT'),
                            Token('ERROR', error='yyy')]).errors,
                     ('xxx', 'yyy'))
        assert_equal(Error([Token('ERROR', error=e) for e in '0123456789']).errors,
                     tuple('0123456789'))

    def test_model_error(self):
        model = get_model('''\
*** Invalid ***
*** Settings ***
Invalid
Documentation
''', data_only=True)
        inv_header = (
            "Unrecognized section header '*** Invalid ***'. Valid sections: "
            "'Settings', 'Variables', 'Test Cases', 'Tasks', 'Keywords' and 'Comments'."
        )
        inv_setting = "Non-existing setting 'Invalid'."
        expected = File([
            InvalidSection(
                header=SectionHeader(
                    [Token('INVALID HEADER', '*** Invalid ***', 1, 0, inv_header)]
                )

            ),
            SettingSection(
                header=SectionHeader([
                    Token('SETTING HEADER', '*** Settings ***', 2, 0)
                ]),
                body=[
                    Error([Token('ERROR', 'Invalid', 3, 0, inv_setting)]),
                    Documentation([Token('DOCUMENTATION', 'Documentation', 4, 0)])
                ]
            )
        ])
        assert_model(model, expected)

    def test_model_error_with_fatal_error(self):
        model = get_resource_model('''\
*** Test Cases ***
''', data_only=True)
        inv_testcases = "Resource file with 'Test Cases' section is invalid."
        expected = File([
            InvalidSection(
                header=SectionHeader(
                    [Token('INVALID HEADER', '*** Test Cases ***', 1, 0, inv_testcases)])
            )
        ])
        assert_model(model, expected)

    def test_model_error_with_error_and_fatal_error(self):
        model = get_resource_model('''\
*** Invalid ***
*** Settings ***
Invalid
Documentation
*** Test Cases ***
''', data_only=True)
        inv_header = (
            "Unrecognized section header '*** Invalid ***'. Valid sections: "
            "'Settings', 'Variables', 'Keywords' and 'Comments'."
        )
        inv_setting = "Non-existing setting 'Invalid'."
        inv_testcases = "Resource file with 'Test Cases' section is invalid."
        expected = File([
            InvalidSection(
                header=SectionHeader(
                    [Token('INVALID HEADER', '*** Invalid ***', 1, 0, inv_header)]
                )
            ),
            SettingSection(
                header=SectionHeader([
                    Token('SETTING HEADER', '*** Settings ***', 2, 0)
                ]),
                body=[
                    Error([Token('ERROR', 'Invalid', 3, 0, inv_setting)]),
                    Documentation([Token('DOCUMENTATION', 'Documentation', 4, 0)]),
                ]
            ),
            InvalidSection(
                header=SectionHeader(
                    [Token('INVALID HEADER', '*** Test Cases ***', 5, 0, inv_testcases)]
                )
            ),
        ])
        assert_model(model, expected)

    def test_set_errors_explicitly(self):
        error = Error([])
        error.errors = ('explicitly set', 'errors')
        assert_equal(error.errors, ('explicitly set', 'errors'))
        error.tokens = [Token('ERROR', error='normal error'),]
        assert_equal(error.errors, ('normal error',
                                    'explicitly set', 'errors'))
        error.errors = ['errors', 'as', 'list']
        assert_equal(error.errors, ('normal error',
                                    'errors', 'as', 'list'))


class TestModelVisitors(unittest.TestCase):

    def test_ast_NodeVisitor(self):

        class Visitor(ast.NodeVisitor):

            def __init__(self):
                self.test_names = []
                self.kw_names = []

            def visit_TestCaseName(self, node):
                self.test_names.append(node.name)

            def visit_KeywordName(self, node):
                self.kw_names.append(node.name)

            def visit_Block(self, node):
                raise RuntimeError('Should not be executed.')

            def visit_Statement(self, node):
                raise RuntimeError('Should not be executed.')

        visitor = Visitor()
        visitor.visit(get_model(DATA))
        assert_equal(visitor.test_names, ['Example'])
        assert_equal(visitor.kw_names, ['Keyword'])

    def test_ModelVisitor(self):

        class Visitor(ModelVisitor):

            def __init__(self):
                self.test_names = []
                self.kw_names = []
                self.blocks = []
                self.statements = []

            def visit_TestCaseName(self, node):
                self.test_names.append(node.name)
                self.visit_Statement(node)

            def visit_KeywordName(self, node):
                self.kw_names.append(node.name)
                self.visit_Statement(node)

            def visit_Block(self, node):
                self.blocks.append(type(node).__name__)
                self.generic_visit(node)

            def visit_Statement(self, node):
                self.statements.append(node.type)

        visitor = Visitor()
        visitor.visit(get_model(DATA))
        assert_equal(visitor.test_names, ['Example'])
        assert_equal(visitor.kw_names, ['Keyword'])
        assert_equal(visitor.blocks,
                     ['ImplicitCommentSection', 'TestCaseSection', 'TestCase',
                      'KeywordSection', 'Keyword'])
        assert_equal(visitor.statements,
                     ['EOL', 'TESTCASE HEADER', 'EOL', 'TESTCASE NAME',
                      'COMMENT', 'KEYWORD', 'EOL', 'EOL', 'KEYWORD HEADER',
                      'COMMENT', 'KEYWORD NAME', 'ARGUMENTS', 'KEYWORD',
                      'RETURN STATEMENT'])

    def test_ast_NodeTransformer(self):

        class Transformer(ast.NodeTransformer):

            def visit_Tags(self, node):
                return None

            def visit_TestCaseSection(self, node):
                self.generic_visit(node)
                node.body.append(
                    TestCase(TestCaseName([Token('TESTCASE NAME', 'Added'),
                                           Token('EOL', '\n')]))
                )
                return node

            def visit_TestCase(self, node):
                self.generic_visit(node)
                return node if node.name != 'REMOVE' else None

            def visit_TestCaseName(self, node):
                name_token = node.get_token(Token.TESTCASE_NAME)
                name_token.value = name_token.value.upper()
                return node

            def visit_Block(self, node):
                raise RuntimeError('Should not be executed.')

            def visit_Statement(self, node):
                raise RuntimeError('Should not be executed.')

        model = get_model('''\
*** Test Cases ***
Example
    [Tags]    to be removed
Remove
''')
        Transformer().visit(model)
        expected = File(sections=[
            TestCaseSection(
                header=SectionHeader([
                    Token('TESTCASE HEADER', '*** Test Cases ***', 1, 0),
                    Token('EOL', '\n', 1, 18)
                ]),
                body=[
                    TestCase(TestCaseName([
                        Token('TESTCASE NAME', 'EXAMPLE', 2, 0),
                        Token('EOL', '\n', 2, 7)
                    ]), errors= ('Test cannot be empty.',)),
                    TestCase(TestCaseName([
                        Token('TESTCASE NAME', 'Added'),
                        Token('EOL', '\n')
                    ]))
                ]
            )
        ])
        assert_model(model, expected)

    def test_ModelTransformer(self):

        class Transformer(ModelTransformer):

            def visit_SectionHeader(self, node):
                return node

            def visit_TestCaseName(self, node):
                return node

            def visit_Statement(self, node):
                return None

            def visit_Block(self, node):
                self.generic_visit(node)
                if hasattr(node, 'header'):
                    for token in node.header.data_tokens:
                        token.value = token.value.upper()
                return node

        model = get_model('''\
*** Test Cases ***
Example
    [Tags]    to be removed
    To be removed
''')
        Transformer().visit(model)
        expected = File(sections=[
            TestCaseSection(
                header=SectionHeader([
                    Token('TESTCASE HEADER', '*** TEST CASES ***', 1, 0),
                    Token('EOL', '\n', 1, 18)
                ]),
                body=[
                    TestCase(TestCaseName([
                        Token('TESTCASE NAME', 'EXAMPLE', 2, 0),
                        Token('EOL', '\n', 2, 7)
                    ])),
                ]
            )
        ])
        assert_model(model, expected)

    def test_visit_Return(self):
        class VisitReturn(ModelVisitor):
            def visit_Return(self, node):
                self.node = node

        for cls in Return, ReturnSetting:
            visitor = VisitReturn()
            ret = cls.from_params(())
            visitor.visit(ret)
            assert_equal(visitor.node, ret)

    def test_visit_ReturnSetting(self):
        class VisitReturnSetting(ModelVisitor):
            def visit_ReturnSetting(self, node):
                self.node = node

        for cls in Return, ReturnSetting:
            visitor = VisitReturnSetting()
            ret = cls.from_params(())
            visitor.visit(ret)
            assert_equal(visitor.node, ret)


class TestLanguageConfig(unittest.TestCase):

    def test_config(self):
        model = get_model('''\
language: fi
language: bad
language: bad    but ignored
language: de     # ok
*** Einstellungen ***
Dokumentaatio    Header is de and setting is fi.
''')
        expected = File(
            languages=('fi', 'de'),
            sections=[
                ImplicitCommentSection(body=[
                    Config([
                        Token('CONFIG', 'language: fi', 1, 0),
                        Token('EOL', '\n', 1, 12)
                    ]),
                    Error([
                        Token('ERROR', 'language: bad', 2, 0,
                              "Invalid language configuration: Language 'bad' "
                              "not found nor importable as a language module."),
                        Token('EOL', '\n', 2, 13)
                    ]),
                    Comment([
                        Token('COMMENT', 'language: bad', 3, 0),
                        Token('SEPARATOR', '    ', 3, 13),
                        Token('COMMENT', 'but ignored', 3, 17),
                        Token('EOL', '\n', 3, 28)
                    ]),
                    Config([
                        Token('CONFIG', 'language: de', 4, 0),
                        Token('SEPARATOR', '     ', 4, 12),
                        Token('COMMENT', '# ok', 4, 17),
                        Token('EOL', '\n', 4, 21)
                    ]),
                ]),
                SettingSection(
                    header=SectionHeader([
                        Token('SETTING HEADER', '*** Einstellungen ***', 5, 0),
                        Token('EOL', '\n', 5, 21)
                    ]),
                    body=[
                        Documentation([
                            Token('DOCUMENTATION', 'Dokumentaatio', 6, 0),
                            Token('SEPARATOR', '    ', 6, 13),
                            Token('ARGUMENT', 'Header is de and setting is fi.', 6, 17),
                            Token('EOL', '\n', 6, 48)
                        ])
                    ]
                )
            ]
        )
        assert_model(model, expected)


if __name__ == '__main__':
    unittest.main()



class InternalLinking:
    u"""Library for testing libdoc's internal linking.

    = Linking to sections =

    - `Introduction`
    - `Library INTROduction`
    - `importing`
    - `Library Importing`
    - `ShortCuts`
    - `Keywords`

    = Linking to keywords =

    - `Keyword`
    - `secoNd kEywoRD`

    = Linking to headers =

    == First level header can be linked ==

    - `linking to headers`
    - `first = level =`

    == Other levels can be linked too ==

    - `Second level`
    - `third level`

    =   First = Level = =

    == Second level ==

    === Third level ===

    = Escaping =

    == Percent encoding: !"#%/()=?|+-_.!~*'() ==

    == HTML entities: &<> ==

    == Non-ASCII: \xe4\u2603 ==

    = Formatting =

    Non-matching `backticks` just get special formatting.
    """

    def __init__(self, argument=None):
        """Importing. See `introduction`, `formatting` and `keyword` for details."""

    def keyword(self):
        """First keyword here. See also `Importing` and `Second Keyword`."""

    def second_keyword(self, arg):
        """We got `arg`. And have `no link`. Except to `Second LEVEL`.

        = Not linkable =

        We are `linking to headers` and `shortcuts` but not to `not linkable`.
        """

    def escaping(self):
        u"""Escaped links:
        - `Percent encoding: !"#%/()=?|+-_.!~*'()`
        - `HTML entities: &<>`
        - `Non-ASCII: \xe4\u2603`
        """



import unittest

from robot.variables import Variables
from robot.errors import DataError, VariableError
from robot.utils.asserts import assert_equal, assert_raises


SCALARS = ['${var}', '${  v A  R }']
LISTS = ['@{var}', '@{  v A  R }']
NOKS = ['var', '$var', '${var', '${va}r', '@{va}r', '@var', '%{var}', ' ${var}',
        '@{var} ', '\\${var}', '\\\\${var}', 42, None, ['${var}'], DataError]


class PythonObject:
    def __init__(self, a, b):
        self.a = a
        self.b = b
    def __getitem__(self, index):
        return (self.a, self.b)[index]
    def __str__(self):
        return '(%s, %s)' % (self.a, self.b)
    def __len__(self):
        return 2
    __repr__ = __str__


class TestVariables(unittest.TestCase):

    def setUp(self):
        self.varz = Variables()

    def test_set(self):
        value = ['value']
        for var in SCALARS + LISTS:
            self.varz[var] = value
            assert_equal(self.varz[var], value)
            assert_equal(self.varz[var.lower().replace(' ', '')], value)
            self.varz.clear()

    def test_set_invalid(self):
        for var in NOKS:
            assert_raises(DataError, self.varz.__setitem__, var, 'value')

    def test_set_scalar(self):
        for var in SCALARS:
            for value in ['string', '', 10, ['hi', 'u'], ['hi', 2],
                          {'a': 1, 'b': 2}, self, None, unittest.TestCase]:
                self.varz[var] = value
                assert_equal(self.varz[var], value)

    def test_set_list(self):
        for var in LISTS:
            for value in [[], [''], ['str'], [10], ['hi', 'u'], ['hi', 2],
                          [{'a': 1, 'b': 2}, self, None]]:
                self.varz[var] = value
                assert_equal(self.varz[var], value)
                self.varz.clear()

    def test_replace_scalar(self):
        self.varz['${foo}'] = 'bar'
        self.varz['${a}'] = 'ari'
        for inp, exp in [('${foo}', 'bar'),
                         ('${a}', 'ari'),
                         (r'$\{a}', '${a}'),
                         ('', ''),
                         ('hii', 'hii'),
                         ("Let's go to ${foo}!", "Let's go to bar!"),
                         ('${foo}ba${a}-${a}', 'barbaari-ari')]:
            assert_equal(self.varz.replace_scalar(inp), exp)

    def test_replace_list(self):
        self.varz['@{L}'] = ['v1', 'v2']
        self.varz['@{E}'] = []
        self.varz['@{S}'] = ['1', '2', '3']
        for inp, exp in [(['@{L}'], ['v1', 'v2']),
                         (['@{L}', 'v3'], ['v1', 'v2', 'v3']),
                         (['v0', '@{L}', '@{E}', 'v${S}[2]'], ['v0', 'v1', 'v2', 'v3']),
                         ([], []),
                         (['hi u', 'hi 2', 3], ['hi u','hi 2', 3])]:
            assert_equal(self.varz.replace_list(inp), exp)

    def test_replace_list_in_scalar_context(self):
        self.varz['@{list}'] = ['v1', 'v2']
        assert_equal(self.varz.replace_list(['@{list}']), ['v1', 'v2'])
        assert_equal(self.varz.replace_list(['-@{list}-']), ["-['v1', 'v2']-"])

    def test_replace_list_item(self):
        self.varz['@{L}'] = ['v0', 'v1']
        assert_equal(self.varz.replace_list(['${L}[0]']), ['v0'])
        assert_equal(self.varz.replace_scalar('${L}[1]'), 'v1')
        assert_equal(self.varz.replace_scalar('-${L}[0]${L}[1]${L}[0]-'), '-v0v1v0-')
        self.varz['${L2}'] = ['v0', ['v11', 'v12']]
        assert_equal(self.varz.replace_list(['${L2}[0]']), ['v0'])
        assert_equal(self.varz.replace_list(['${L2}[1]']), [['v11', 'v12']])
        assert_equal(self.varz.replace_scalar('${L2}[0]'), 'v0')
        assert_equal(self.varz.replace_scalar('${L2}[1]'), ['v11', 'v12'])
        assert_equal(self.varz.replace_list(['${L}[0]', '@{L2}[1]']), ['v0', 'v11', 'v12'])

    def test_replace_dict_item(self):
        self.varz['&{D}'] = {'a': 1, 2: 'b', 'nested': {'a': 1}}
        assert_equal(self.varz.replace_scalar('${D}[a]'), 1)
        assert_equal(self.varz.replace_scalar('${D}[${2}]'), 'b')
        assert_equal(self.varz.replace_scalar('${D}[nested][a]'), 1)
        assert_equal(self.varz.replace_scalar('${D}[nested]'), {'a': 1})
        assert_equal(self.varz.replace_scalar('&{D}[nested]'), {'a': 1})

    def test_replace_non_strings(self):
        self.varz['${d}'] = {'a': 1, 'b': 2}
        self.varz['${n}'] = None
        assert_equal(self.varz.replace_scalar('${d}'), {'a': 1, 'b': 2})
        assert_equal(self.varz.replace_scalar('${n}'), None)

    def test_replace_non_strings_inside_string(self):
        class Example:
            def __str__(self):
                return 'Hello'
        self.varz['${h}'] = Example()
        self.varz['${w}'] = 'world'
        res = self.varz.replace_scalar('Another "${h} ${w}" example')
        assert_equal(res, 'Another "Hello world" example')

    def test_replace_list_item_invalid(self):
        self.varz['@{L}'] = ['v0', 'v1', 'v3']
        for inv in ['@{L}[3]', '@{NON}[0]', '@{L[2]}']:
            assert_raises(VariableError, self.varz.replace_list, [inv])

    def test_replace_non_existing_list(self):
        assert_raises(VariableError, self.varz.replace_list, ['${nonexisting}'])

    def test_replace_non_existing_scalar(self):
        assert_raises(VariableError, self.varz.replace_scalar, '${nonexisting}')

    def test_replace_non_existing_string(self):
        assert_raises(VariableError, self.varz.replace_string, '${nonexisting}')

    def test_non_string_input(self):
        for item in [1, False, None, [], (), {}, object]:
            assert_equal(self.varz.replace_list([item]), [item])
            assert_equal(self.varz.replace_scalar(item), item)
            assert_equal(self.varz.replace_string(item), str(item))

    def test_replace_escaped(self):
        self.varz['${foo}'] = 'bar'
        for inp, exp in [(r'\${foo}', r'${foo}'),
                         (r'\\${foo}', r'\bar'),
                         (r'\\\${foo}', r'\${foo}'),
                         (r'\\\\${foo}', r'\\bar'),
                         (r'\\\\\${foo}', r'\\${foo}')]:
            assert_equal(self.varz.replace_scalar(inp), exp)

    def test_variables_in_value(self):
        self.varz['${exists}'] = 'Variable exists but is still not replaced'
        self.varz['${test}'] = '${exists} & ${does_not_exist}'
        assert_equal(self.varz['${test}'], '${exists} & ${does_not_exist}')
        self.varz['@{test}'] = ['${exists}', '&', '${does_not_exist}']
        assert_equal(self.varz['@{test}'], '${exists} & ${does_not_exist}'.split())

    def test_variable_as_object(self):
        obj = PythonObject('a', 1)
        self.varz['${obj}'] = obj
        assert_equal(self.varz['${obj}'], obj)
        expected = ['Some text here %s and %s there' % (obj, obj)]
        actual = self.varz.replace_list(['Some text here ${obj} and ${obj} there'])
        assert_equal(actual, expected)

    def test_extended_variables(self):
        # Extended variables are vars like ${obj.name} when we have var ${obj}
        obj = PythonObject('a', [1, 2, 3])
        dic = {'a': 1, 'o': obj}
        self.varz['${obj}'] = obj
        self.varz['${dic}'] = dic
        assert_equal(self.varz.replace_scalar('${obj.a}'), 'a')
        assert_equal(self.varz.replace_scalar('${obj.b}'), [1, 2, 3])
        assert_equal(self.varz.replace_scalar('${obj.b[0]}-${obj.b[1]}'), '1-2')
        assert_equal(self.varz.replace_scalar('${dic["a"]}'), 1)
        assert_equal(self.varz.replace_scalar('${dic["o"]}'), obj)
        assert_equal(self.varz.replace_scalar('-${dic["o"].b[2]}-'), '-3-')

    def test_space_is_not_ignored_after_newline_in_extend_variable_syntax(self):
        self.varz['${x}'] = 'test string'
        self.varz['${lf}'] = '\\n'
        self.varz['${lfs}'] = '\\n '
        for inp, exp in [('${x.replace(" ", """\\n""")}', 'test\nstring'),
                         ('${x.replace(" ", """\\n """)}', 'test\n string'),
                         ('${x.replace(" ", """${lf}""")}', 'test\nstring'),
                         ('${x.replace(" ", """${lfs}""")}', 'test\n string')]:
            assert_equal(self.varz.replace_scalar(inp), exp)

    def test_escaping_with_extended_variable_syntax(self):
        self.varz['${p}'] = 'c:\\temp'
        assert self.varz['${p}'] == 'c:\\temp'
        assert_equal(self.varz.replace_scalar('${p + "\\\\foo.txt"}'),
                     'c:\\temp\\foo.txt')

    def test_internal_variables(self):
        # Internal variables are variables like ${my${name}}
        self.varz['${name}'] = 'name'
        self.varz['${my name}'] = 'value'
        assert_equal(self.varz.replace_scalar('${my${name}}'), 'value')
        self.varz['${whos name}'] = 'my'
        assert_equal(self.varz.replace_scalar('${${whos name} ${name}}'), 'value')
        assert_equal(self.varz.replace_scalar('${${whos${name}}${name}}'), 'value')
        self.varz['${my name}'] = [1, 2, 3]
        assert_equal(self.varz.replace_scalar('${${whos${name}}${name}}'), [1, 2, 3])
        assert_equal(self.varz.replace_scalar('- ${${whos${name}}${name}} -'), '- [1, 2, 3] -')

    def test_math_with_internal_vars(self):
        assert_equal(self.varz.replace_scalar('${${1}+${2}}'), 3)
        assert_equal(self.varz.replace_scalar('${${1}-${2}}'), -1)
        assert_equal(self.varz.replace_scalar('${${1}*${2}}'), 2)
        assert_equal(self.varz.replace_scalar('${${1}//${2}}'), 0)

    def test_math_with_internal_vars_with_spaces(self):
        assert_equal(self.varz.replace_scalar('${${1} + ${2.5}}'), 3.5)
        assert_equal(self.varz.replace_scalar('${${1} - ${2} + 1}'), 0)
        assert_equal(self.varz.replace_scalar('${${1} * ${2} - 1}'), 1)
        assert_equal(self.varz.replace_scalar('${${1} / ${2.0}}'), 0.5)

    def test_math_with_internal_vars_does_not_work_if_first_var_is_float(self):
        assert_raises(VariableError, self.varz.replace_scalar, '${${1.1}+${2}}')
        assert_raises(VariableError, self.varz.replace_scalar, '${${1.1} - ${2}}')
        assert_raises(VariableError, self.varz.replace_scalar, '${${1.1} * ${2}}')
        assert_raises(VariableError, self.varz.replace_scalar, '${${1.1}/${2}}')

    def test_list_variable_as_scalar(self):
        self.varz['@{name}'] = exp = ['spam', 'eggs']
        assert_equal(self.varz.replace_scalar('${name}'), exp)
        assert_equal(self.varz.replace_list(['${name}', 42]), [exp, 42])
        assert_equal(self.varz.replace_string('${name}'), str(exp))

    def test_copy(self):
        varz = Variables()
        varz['${foo}'] = 'bar'
        copy = varz.copy()
        assert_equal(copy['${foo}'], 'bar')

    def test_ignore_error(self):
        v = Variables()
        v['${X}'] = 'x'
        v['@{Y}'] = [1, 2, 3]
        for item in ['${foo}', 'foo${bar}', '${foo}', '@{zap}', '${Y}[7]',
                     '${inv', '${{inv}', '${var}[inv', '${var}[key][inv']:
            x_at_end = 'x' if (item.count('{') == item.count('}') and
                               item.count('[') == item.count(']')) else '${x}'
            assert_equal(v.replace_string(item, ignore_errors=True), item)
            assert_equal(v.replace_string('${x}'+item+'${x}', ignore_errors=True),
                         'x' + item + x_at_end)
            assert_equal(v.replace_scalar(item, ignore_errors=True), item)
            assert_equal(v.replace_scalar('${x}'+item+'${x}', ignore_errors=True),
                         'x' + item + x_at_end)
            assert_equal(v.replace_list([item], ignore_errors=True), [item])
            assert_equal(v.replace_list(['${X}', item, '@{Y}'], ignore_errors=True),
                         ['x', item, 1, 2, 3])
            assert_equal(v.replace_list(['${x}'+item+'${x}', '@{NON}'], ignore_errors=True),
                         ['x' + item + x_at_end, '@{NON}'])

    def test_sequence_subscript(self):
        sequences = (
            [42, 'my', 'name'],
            (42, ['foo', 'bar'], 'name'),
            'abcDEF123#@$',
            b'abcDEF123#@$',
            bytearray(b'abcDEF123#@$'),
        )
        for var in sequences:
            self.varz['${var}'] = var
            assert_equal(self.varz.replace_scalar('${var}[0]'), var[0])
            assert_equal(self.varz.replace_scalar('${var}[-2]'), var[-2])
            assert_equal(self.varz.replace_scalar('${var}[::2]'), var[::2])
            assert_equal(self.varz.replace_scalar('${var}[1::2]'), var[1::2])
            assert_equal(self.varz.replace_scalar('${var}[1:-3:2]'), var[1:-3:2])
            assert_raises(VariableError, self.varz.replace_scalar, '${var}[0][1]')

    def test_dict_subscript(self):
        a_key = (42, b'key')
        var = {'foo': 'bar', 42: [4, 2], 'name': b'my-name', a_key: {4: 2}}
        self.varz['${a_key}'] = a_key
        self.varz['${var}'] = var
        assert_equal(self.varz.replace_scalar('${var}[foo][-1]'), var['foo'][-1])
        assert_equal(self.varz.replace_scalar('${var}[${42}][-1]'), var[42][-1])
        assert_equal(self.varz.replace_scalar('${var}[name][:3]'), var['name'][:3])
        assert_equal(self.varz.replace_scalar('${var}[${a_key}][${4}]'), var[a_key][4])
        assert_raises(VariableError, self.varz.replace_scalar, '${var}[1]')
        assert_raises(VariableError, self.varz.replace_scalar, '${var}[42:]')
        assert_raises(VariableError, self.varz.replace_scalar, '${var}[nonex]')

    def test_custom_class_subscriptable_like_sequence(self):
        # the two class attributes are accessible via indices 0 and 1
        # slicing should be supported here as well
        bytes_key = b'my'
        var = PythonObject([1, 2, 3, 4, 5], {bytes_key: 'myname'})
        self.varz['${bytes_key}'] = bytes_key
        self.varz['${var}'] = var
        assert_equal(self.varz.replace_scalar('${var}[${0}][2::2]'), [3, 5])
        assert_equal(self.varz.replace_scalar('${var}[0][2::2]'), [3, 5])
        assert_equal(self.varz.replace_scalar('${var}[1][${bytes_key}][2:]'), 'name')
        assert_equal(self.varz.replace_scalar('${var}\\[1]'), str(var) + '[1]')
        assert_equal(self.varz.replace_scalar('${var}[:][0][4]'), var[:][0][4])
        assert_equal(self.varz.replace_scalar('${var}[:-2]'), var[:-2])
        assert_equal(self.varz.replace_scalar('${var}[:7:-2]'), var[:7:-2])
        assert_equal(self.varz.replace_scalar('${var}[2::]'), ())
        assert_raises(VariableError, self.varz.replace_scalar, '${var}[${2}]')
        assert_raises(VariableError, self.varz.replace_scalar, '${var}[${bytes_key}]')

    def test_non_subscriptable(self):
        assert_raises(VariableError, self.varz.replace_scalar, '${1}[1]')


if __name__ == '__main__':
    unittest.main()
